package gve.calc.formula;



import awt.*;

import java.awt.*; 

import awt.Button;

import java.awt.event.*;

import gve.calc.logic.*;

import java.applet.Applet;



/** Binary boolean operators */

public class OperatorBoolean extends InfixBinaryOp 

{	private byte kind;



	public static final byte AND = 0;

	public static final byte XOR = 1;

	public static final byte OR = 2;

	public static final byte IMPLIES = 3;

	public static final byte EQUIV = 4;



	private static String [] names = { "&","xor","or","=>","<=>" };

	private static int [] pri = { 1200,1100,1000,900,900 };



	public OperatorBoolean(String name,Part l,Part r) 

	{	super(l,r);

		for (int i = names.length-1; i>=0; i--) 

		{	if (names[i].equals(name)) 

			{	kind = (byte)i;

				break;

			}

		}

	}



	public OperatorBoolean(byte kind,Part l,Part r) 

	{	super(l,r);

		this.kind = kind;

	}



	public String getName() { return names[kind]; }

	public int getPri() { return pri[kind]; }

	public int getLeftPri() { return pri[kind] - 10; }

	public int getRightPri() { return pri[kind] + 10; }

	public byte getKind() { return kind; }



	public Part evaluate(Evaluator ev) 

	{	// PENDING: shortcut L->R evaluatie bij sommige operatoren

		Object evalL = left.evaluate(ev);

		Object evalR = right.evaluate(ev);

		if (evalL instanceof Boolean && evalR instanceof Boolean) 

		{	boolean l = ((Boolean)evalL).getBooleanValue();

			boolean r = ((Boolean)evalR).getBooleanValue();

			switch(kind) 

			{	case AND: return new Boolean(l & r);

				case XOR: return new Boolean(l != r);

				case OR: return new Boolean(l | r);

				case IMPLIES: return new Boolean(!l | r);

				case EQUIV: return new Boolean(l == r);

			}

		}

		return new OperatorBoolean(kind,(Part)evalL,(Part)evalR);

	}



	public boolean same(Object o) 

	{	o = Brackets.unbracket(o);

		if (!(o instanceof OperatorBoolean)) return false;

		OperatorBoolean b = (OperatorBoolean)o;

		return b.left.same(left) && b.right.same(right) && b.kind==kind;

	}



	public void write(java.io.Writer w) throws java.io.IOException 

	{	super.write(w);

		w.write(kind+"\n");

	}



	public static Part read(java.io.BufferedReader r) throws java.io.IOException,

		ClassNotFoundException,NoSuchMethodException,

		java.lang.reflect.InvocationTargetException,IllegalAccessException

	{	Part left = Part.read(r);

		Part right = Part.read(r);

		String line = r.readLine();

		byte kind = (byte)Integer.parseInt(line);

		return new OperatorBoolean(kind,left,right);

	}



	public Component createView(FormulaView view) 

	{	return new OperatorBooleanView(this,view,new OperatorBooleanSymbolView(kind,getName()));

	}



	static public void herschrijf(Part p, Formula f) 

	{	if (p instanceof OperatorBoolean) 

		{	OperatorBoolean ob = (OperatorBoolean)p;

			p = ob.rewrite(f);

		}

		int i = 0;

		while (true)

 		{	Part kind = p.getChild(i);

			if (kind == null) break;

			herschrijf(kind,f);

		}

	}



	public Part rewrite(Formula f)

 	{	OperatorBoolean operator = this; 

		OperatorModel nieuw = new OperatorModel("|>",new Identifier(""),new Identifier(""));

		Part nieuwetop = (Part)nieuw;



		//herschrijven naar not en &

		Identifier dummy = new Identifier("");

		switch(operator.getKind()) 

		{			

		case OperatorBoolean.OR:		// a or b == not(not a&not b)

		    {	f.replace(operator,dummy);

			Part left = operator.left, right = operator.right;

			int notpri = new OperatorNot(new Identifier("")).getDownPri();

			if (left instanceof InfixBinaryOp && ((InfixBinaryOp)left).getPri()<notpri)

				left = new Brackets(left);

			if (right instanceof InfixBinaryOp && ((InfixBinaryOp)right).getPri()<notpri)

				right = new Brackets(right);

			nieuwetop = new OperatorNot(new Brackets(

				new OperatorBoolean("&",new OperatorNot(left),new OperatorNot(right))

			));

			f.replace(dummy,nieuwetop);

		    }

			break;



		case OperatorBoolean.XOR:  		// a xor b == (not(a & b)) & (not(not a & not b))

		  { 	f.replace(operator,dummy);

			Part left = operator.left, right = operator.right;

			Part left1 = (Part)left.clone();				

			Part left2 = (Part)left.clone();				

			Part right1 = (Part)right.clone();				

			Part right2 = (Part)right.clone();

			int notpri = new OperatorNot(new Identifier("")).getDownPri();

			int andLpri = getLeftPri();

			int andRpri = getRightPri();

			if (left instanceof InfixBinaryOp && ((InfixBinaryOp)left).getPri()<andLpri)

				left1 = new Brackets(left);

			if (right instanceof InfixBinaryOp && ((InfixBinaryOp)right).getPri()<andRpri)

				right1 = new Brackets(right);

			if (left instanceof InfixBinaryOp && ((InfixBinaryOp)left).getPri()<notpri)

				left2 = new Brackets(left);

			if (right instanceof InfixBinaryOp && ((InfixBinaryOp)right).getPri()<notpri)

				right2 = new Brackets(right);

			nieuwetop = new OperatorBoolean("&",new Brackets(

				new OperatorNot(new Brackets(new OperatorBoolean("&",left1,right1)))),

				new Brackets(new OperatorNot(new Brackets(

				new OperatorBoolean("&",new OperatorNot(left2),new OperatorNot(right2))))));

			f.replace(dummy,nieuwetop);

		   }

			break;



		case OperatorBoolean.IMPLIES:	// a=>b == not a or b == not(a & not b)

		    {	f.replace(operator,dummy);

		    	// STRANGE: if we don't clone the left Part, the syntax tree comes out OK,

			// but it's not shown: the user sees "not(&not rightpart)" !!

			Part left = (Part)operator.left.clone(), right = operator.right;

			int notpri = new OperatorNot(new Identifier("")).getDownPri();

			int andLpri = getLeftPri();

			if (left instanceof InfixBinaryOp && ((InfixBinaryOp)left).getPri()<andLpri)

				left = new Brackets(left);

			if (right instanceof InfixBinaryOp && ((InfixBinaryOp)right).getPri()<notpri)

				right = new Brackets(right);

			nieuwetop = new OperatorNot(new Brackets(

				new OperatorBoolean("&",left,new OperatorNot(right))

			));

			f.replace(dummy,nieuwetop);

		    }

			break;



		case OperatorBoolean.EQUIV:  	// a<->b == ((not(a & not b)) & (not(b & not a)))

		    {	f.replace(operator,dummy);

			Part left = (Part)operator.left.clone(), right = operator.right;

			Part left1 = (Part)left.clone();				

			Part left2 = (Part)right.clone();				

			Part right1 = (Part)right.clone();				

			Part right2 = (Part)left.clone();

			int notpri = new OperatorNot(new Identifier("")).getDownPri();

			int andLpri = getLeftPri();

			if (left instanceof InfixBinaryOp && ((InfixBinaryOp)left).getPri()<andLpri)

				left1 = new Brackets(left);

			if (right instanceof InfixBinaryOp && ((InfixBinaryOp)right).getPri()<notpri)

				right1 = new Brackets(right);	

			if (right instanceof InfixBinaryOp && ((InfixBinaryOp)right).getPri()<andLpri)

				left2 = new Brackets(right);	

			if (left instanceof InfixBinaryOp && ((InfixBinaryOp)left).getPri()<notpri)

				right2 = new Brackets(left);

			nieuwetop = new OperatorBoolean("&",new Brackets(new OperatorNot(new Brackets(

				new OperatorBoolean("&",left1,new OperatorNot(right1))))),new Brackets(new OperatorNot(new Brackets(

				new OperatorBoolean("&",left2,new OperatorNot(right2))))));

			f.replace(dummy,nieuwetop);

		   }

			break;

		}

	return nieuwetop;			

	}



	public void reduce (Formula f) 

	{	OperatorBoolean operator = this;

		OperatorModel nieuw = new OperatorModel("|>",new Identifier (""),new Identifier(""));

				

		//reduceren

		switch(operator.getKind())

		{          	

		case OperatorBoolean.AND:

	        	    {   	Part p = operator;

		 	while (p.parent instanceof OperatorComma) p = p.parent;

			if (p.parent instanceof OperatorModel && ((OperatorModel)p.parent).getKind()==OperatorModel.TRIANGLE) 

			{	OperatorModel mod = (OperatorModel)p.parent;

					if (p == mod.right) 		// Reduceren aan de rechterkant

				{	// Add double branch

					Identifier dummy = new Identifier("");

					Part rightlist,leftlist;

					OperatorModel leftbranch = new OperatorModel("|>",(Part)mod.left.clone(),

						leftlist=OperatorComma.copyListExcept(mod.right,operator));

					OperatorModel rightbranch = new OperatorModel("|>",(Part)mod.left.clone(),

						rightlist=OperatorComma.copyListExcept(mod.right,operator));

					f.replace(mod,new TableauPart(dummy,leftbranch,rightbranch));

					f.replace(dummy,mod);

					OperatorComma.addToFrontOfList(f,leftlist,(Part)operator.left.clone());

					OperatorComma.addToFrontOfList(f,rightlist,(Part)operator.right.clone());

					leftbranch.reduceer(f);

					rightbranch.reduceer(f);

				}

				else if (p == mod.left) 	// Reduceren aan de linkerkant

				{	Identifier dummy = new Identifier("");

					OperatorModel newmod = new OperatorModel("|>",dummy,(Part)mod.right.clone());

					{	// Add branch

						Identifier dummy2 = new Identifier("");

						f.replace(mod,new TableauPart(dummy2,newmod));

						f.replace(dummy2,mod);

					}

					if (mod.left instanceof OperatorComma) 

					{	OperatorComma comma = (OperatorComma)mod.left;

						f.replace(dummy,comma.copyListExcept(operator));

						OperatorComma.addToFrontOfList(f,newmod.left,(Part)operator.right.clone());

						OperatorComma.addToFrontOfList(f,newmod.left,(Part)operator.left.clone());	

						newmod.reduceer(f);

					} 

					else 

					{	f.replace(dummy,new OperatorComma((Part)operator.left.clone(),(Part)operator.right.clone()));

						newmod.reduceer(f);

					}

				} 

			}	

	          	    } 	

		break;

		

		 case OperatorBoolean.OR:

	         	     {   	Part p = operator;

			while (p.parent instanceof OperatorComma) p = p.parent;

			if (p.parent instanceof OperatorModel && ((OperatorModel)p.parent).getKind()==OperatorModel.TRIANGLE) 

			{	OperatorModel mod = (OperatorModel)p.parent;



				if (p == mod.right) 		// Reduceren aan de rechterkant

				{	Identifier dummy = new Identifier("");

					OperatorModel newmod = new OperatorModel("|>",(Part)mod.left.clone(),dummy);

					{	Identifier dummy2 = new Identifier("");

						f.replace(mod,new TableauPart(dummy2,newmod));

						f.replace(dummy2,mod);

					}

					if (mod.right instanceof OperatorComma) 

					{	OperatorComma comma = (OperatorComma)mod.right;

						f.replace(dummy,comma.copyListExcept(operator));

						OperatorComma.addToFrontOfList(f,newmod.right,(Part)operator.left.clone());

						OperatorComma.addToFrontOfList(f,newmod.right,(Part)operator.right.clone());	

						newmod.reduceer(p,f);

					} 

					else 

					{	f.replace(dummy,new OperatorComma((Part)operator.left.clone(),(Part)operator.right.clone()));

						newmod.reduceer(p,f);

					}

				} 

				else if (p == mod.left) 	// Reduceren aan de linkerkant

				{	Identifier dummy = new Identifier("");

					Part rightlist,leftlist;

					OperatorModel leftbranch = new OperatorModel("|>",(Part)mod.right.clone(),

						leftlist=OperatorComma.copyListExcept(mod.left,operator));

					OperatorModel rightbranch = new OperatorModel("|>",(Part)mod.right.clone(),

						rightlist=OperatorComma.copyListExcept(mod.left,operator));

					f.replace(mod,new TableauPart(dummy,leftbranch,rightbranch));

					f.replace(dummy,mod);

					OperatorComma.addToFrontOfList(f,leftlist,(Part)operator.left.clone());

					OperatorComma.addToFrontOfList(f,rightlist,(Part)operator.right.clone());

					leftbranch.reduceer(p,f);

					rightbranch.reduceer(p,f);

				}

			}

	          	    } 	

		break;



		case OperatorBoolean.XOR:

	          	    {   	Part p = operator;

			while (p.parent instanceof OperatorComma) p = p.parent;

			if (p.parent instanceof OperatorModel && ((OperatorModel)p.parent).getKind()==OperatorModel.TRIANGLE) 

			{	OperatorModel mod = (OperatorModel)p.parent;

					if (p == mod.right) 		// Reduceren aan de rechterkant

				{	Identifier dummy = new Identifier("");

					Part rightlistR,leftlistL,leftlistR;

					OperatorModel leftbranch = new OperatorModel("|>",leftlistL=(Part)mod.left.clone(),

						leftlistR=OperatorComma.copyListExcept(mod.right,operator));

					OperatorModel rightbranch = new OperatorModel("|>",(Part)mod.left.clone(),

						rightlistR=OperatorComma.copyListExcept(mod.right,operator));

					f.replace(mod,new TableauPart(dummy,leftbranch,rightbranch));

					f.replace(dummy,mod);

					OperatorComma.addToFrontOfList(f,leftlistL,(Part)operator.left.clone());

					OperatorComma.addToFrontOfList(f,leftlistL,(Part)operator.right.clone());

					OperatorComma.addToFrontOfList(f,rightlistR,(Part)operator.left.clone());

					OperatorComma.addToFrontOfList(f,rightlistR,(Part)operator.right.clone());

					leftbranch.reduceer(p,f);

					rightbranch.reduceer(p,f);

				}

				else if (p == mod.left) 	// Reduceren aan de linkerkant

				{	Identifier dummy = new Identifier("");

					Part rightlistL,rightlistR,leftlistL,leftlistR;

					OperatorModel leftbranch = new OperatorModel("|>",leftlistL=OperatorComma.copyListExcept(mod.left,operator),

						leftlistR=(Part)mod.right.clone());

					OperatorModel rightbranch = new OperatorModel("|>",rightlistL=OperatorComma.copyListExcept(mod.left,operator),

						rightlistR=(Part)mod.right.clone());

					f.replace(mod,new TableauPart(dummy,leftbranch,rightbranch));

					f.replace(dummy,mod);

					OperatorComma.addToFrontOfList(f,leftlistL,(Part)operator.right.clone());

					OperatorComma.addToFrontOfList(f,leftlistR,(Part)operator.left.clone());

					OperatorComma.addToFrontOfList(f,rightlistL,(Part)operator.left.clone());

					OperatorComma.addToFrontOfList(f,rightlistR,(Part)operator.right.clone());

					leftbranch.reduceer(p,f);

					rightbranch.reduceer(p,f);

				}

			} 

		    } 	

		break;

	

		case OperatorBoolean.IMPLIES:

	   	    {   	Part p = operator;

			while (p.parent instanceof OperatorComma) p = p.parent;

			if (p.parent instanceof OperatorModel && ((OperatorModel)p.parent).getKind()==OperatorModel.TRIANGLE)

			 {	OperatorModel mod = (OperatorModel)p.parent;

				if (p == mod.right)		// Reduceren aan de rechterkant

				{	Identifier dummy = new Identifier("");	

					OperatorModel newmod = new OperatorModel("|>",(Part)mod.left.clone(),dummy);

					{	Identifier dummy2 = new Identifier("");

						f.replace(mod,new TableauPart(dummy2,newmod));

						f.replace(dummy2,mod);

					}

					if (mod.right instanceof OperatorComma) 

					{	OperatorComma comma = (OperatorComma)mod.right;

						f.replace(dummy,comma.copyListExcept(operator));

						OperatorComma.addToFrontOfList(f,newmod.left,(Part)operator.left.clone());

						OperatorComma.addToFrontOfList(f,newmod.right,(Part)operator.right.clone());

						newmod.reduceer(p,f);

					} 

					else 

					{	f.replace(dummy,(Part)operator.right.clone());

						OperatorComma.addToFrontOfList(f,newmod.left,(Part)operator.left.clone());

						newmod.reduceer(p,f);

					}

				} 

				else if (p == mod.left) 	// Reduceren aan de linkerkant

				{	Identifier dummy = new Identifier("");

					Part rightlistL,rightlistR,leftlist;

					OperatorModel leftbranch = new OperatorModel("|>",(Part)mod.right.clone(),

						leftlist=OperatorComma.copyListExcept(mod.left,operator));

					OperatorModel rightbranch = new OperatorModel("|>",rightlistL=(Part)mod.right.clone(),

						rightlistR=OperatorComma.copyListExcept(mod.left,operator));

					f.replace(mod,new TableauPart(dummy,leftbranch,rightbranch));

					f.replace(dummy,mod);

					OperatorComma.addToFrontOfList(f,leftlist,(Part)operator.left.clone());

					OperatorComma.addToFrontOfList(f,rightlistL,(Part)operator.right.clone());

					leftbranch.reduceer(p,f);

					rightbranch.reduceer(p,f);

				}

			} 

		    } 	

		break;



		 case OperatorBoolean.EQUIV:

 		    {   	Part p = operator;

			while (p.parent instanceof OperatorComma) p = p.parent;

			if (p.parent instanceof OperatorModel && ((OperatorModel)p.parent).getKind()==OperatorModel.TRIANGLE) 

			{	OperatorModel mod = (OperatorModel)p.parent;

					if (p == mod.right) 		// Reduceren aan de rechterkant

				{	Identifier dummy = new Identifier("");

					Part rightlistL,rightlistR,leftlistL,leftlistR;

					OperatorModel leftbranch = new OperatorModel("|>",leftlistL=(Part)mod.left.clone(),

						leftlistR=OperatorComma.copyListExcept(mod.right,operator));

					OperatorModel rightbranch = new OperatorModel("|>",rightlistL=(Part)mod.left.clone(),

						rightlistR=OperatorComma.copyListExcept(mod.right,operator));

					f.replace(mod,new TableauPart(dummy,leftbranch,rightbranch));

					f.replace(dummy,mod);

					OperatorComma.addToFrontOfList(f,leftlistL,(Part)operator.left.clone());

					OperatorComma.addToFrontOfList(f,leftlistR,(Part)operator.right.clone());

					OperatorComma.addToFrontOfList(f,rightlistL,(Part)operator.right.clone());

					OperatorComma.addToFrontOfList(f,rightlistR,(Part)operator.left.clone());

					leftbranch.reduceer(p,f);

					rightbranch.reduceer(p,f);

				}

				else if(p == mod.left) 	// Reduceren aan de linkerkant

				{	Identifier dummy = new Identifier("");

					Part rightlistL,leftlistL,leftlistR;

					OperatorModel leftbranch = new OperatorModel("|>",leftlistL=OperatorComma.copyListExcept(mod.left,operator),

						leftlistR=(Part)mod.right.clone());

					OperatorModel rightbranch = new OperatorModel("|>",rightlistL=OperatorComma.copyListExcept(mod.left,operator),

						(Part)mod.right.clone());

					f.replace(mod,new TableauPart(dummy,leftbranch,rightbranch));

					f.replace(dummy,mod);

					OperatorComma.addToFrontOfList(f,leftlistR,(Part)operator.left.clone());

					OperatorComma.addToFrontOfList(f,leftlistR,(Part)operator.right.clone());

					OperatorComma.addToFrontOfList(f,rightlistL,(Part)operator.left.clone());				

					OperatorComma.addToFrontOfList(f,rightlistL,(Part)operator.right.clone());		

					leftbranch.reduceer(p,f);

					rightbranch.reduceer(p,f);

				}

		          	} 

		    }

      		break;

		

		}//switch		

	}//reduce 

}//class





class OperatorBooleanSymbolView extends TextSymbolView {

	private byte kind;

	public OperatorBooleanSymbolView(byte kind,String str) {

		super(str);

		this.kind = kind;

	}



	public Dimension getSymbolSize() {

		FontMetrics metrics = getFontMetrics(getFont());

		int w = 0;

		switch (kind) {

		case OperatorBoolean.AND: w = metrics.charWidth('&'); break;

		case OperatorBoolean.XOR: w = metrics.charWidth('O'); break;

		case OperatorBoolean.OR: w = metrics.charWidth('v'); break;

		case OperatorBoolean.IMPLIES:

		case OperatorBoolean.EQUIV: w = metrics.stringWidth("=="); break;

		}

		return new Dimension(w,metrics.getHeight());

	}



	public void paintSymbol(Graphics g) {

		FontMetrics metrics = getFontMetrics(getFont());

		int spatwidth = metrics.charWidth(' ');

		int ascent = metrics.getAscent();

		switch (kind) {

		case OperatorBoolean.AND:

			g.drawString("&",0,ascent); return;

		case OperatorBoolean.XOR:

			paintXor(g,0,0,metrics.charWidth('O'),ascent); break;

		case OperatorBoolean.OR:

			paintOr(g,0,ascent/3,metrics.charWidth('v'),ascent*2/3); break;

		case OperatorBoolean.IMPLIES:

			paintImplies(g,1,0,metrics.stringWidth("==")-2,ascent); break;

		case OperatorBoolean.EQUIV:

			paintEquiv(g,1,0,metrics.stringWidth("==")-2,ascent); break;

		}

	}



	private void paintXor(Graphics g,int x,int y,int w,int h) {

		g.drawOval(x,y,w,h);

		w--; h--;

		g.drawLine(x+w/2,y,x+w/2,y+h);

		g.drawLine(x,y+h/2,x+w,y+h/2);

	}



	private void paintOr(Graphics g,int x,int y,int w,int h) {

		w--; h--;

		g.drawLine(x,y,x+w/2,y+h);

		g.drawLine(x+w/2,y+h,x+w,y);

	}



	private void paintImplies(Graphics g,int x,int y,int w,int h) {

		w--; h--;

		// right arrow

		g.drawLine(x+w*3/4,y,x+w,y+h/2);

		g.drawLine(x+w,y+h/2,x+w*3/4,y+h);

		// two horizontal lines

		g.drawLine(x,y+(h+2)/4,x+w*7/8,y+(h+2)/4);

		g.drawLine(x,y+h*3/4,x+w*7/8,y+h*3/4);

	}



	private void paintEquiv(Graphics g,int x,int y,int w,int h) {

		w--; h--;

		// right arrow

		g.drawLine(x+w*3/4,y,x+w,y+h/2);

		g.drawLine(x+w,y+h/2,x+w*3/4,y+h);

		// two horizontal lines

		g.drawLine(x+w/8,y+(h+2)/4,x+w*7/8,y+(h+2)/4);

		g.drawLine(x+w/8,y+h*3/4,x+w*7/8,y+h*3/4);

		// left arrow

		g.drawLine(x+w/4,y,x,y+h/2);

		g.drawLine(x,y+h/2,x+w/4,y+h);

	}

}

