package test;

import junit.framework.*;
import java.awt.*;
import java.awt.event.KeyEvent;
import gve.calc.*;
import gve.calc.formula.*;

public class EditorTest extends GuiTest {

	public EditorTest(String name) { super(name); }

	public static Test suite() {
		TestSuite suite = new TestSuite();
		suite.addTest(new EditorTest("testSpace1"));
		suite.addTest(new EditorTest("testSpace2"));
		suite.addTest(new EditorTest("testSpace3"));
		suite.addTest(new EditorTest("testDot"));
//		Commented out because this tests the experimental Stack class ...
//		suite.addTest(new EditorTest("test5"));
		suite.addTest(new EditorTest("test6"));
		suite.addTest(new EditorTest("test7"));
		suite.addTest(new EditorTest("test8"));
		suite.addTest(new EditorTest("test9"));
		suite.addTest(new EditorTest("testForall"));
		suite.addTest(new EditorTest("testSemicolon"));
		suite.addTest(new EditorTest("testSpace"));
		suite.addTest(new EditorTest("testQuote"));
		suite.addTest(new EditorTest("testRandom"));
		return suite;
	}

	public void testSpace1() {
//		System.out.println("test 1 (2-mar-2000)");
		// Test: type "notx", split with a space in "not x"
		// and check that no superfluous space has been generated
		// (same as "forallx" in the debug report)
		Formula f;
		Part shouldbe = new OperatorNot(new Identifier("x"));
		canvas.push(f = new Formula("notx"));
		canvas.setActiveFormula(0);
		FormulaView view = (FormulaView)canvas.getComponent(0);
		view.keydn(KeydnListener.Key_RIGHT);
		view.keydn(KeydnListener.Key_RIGHT);
		view.keydn(KeydnListener.Key_RIGHT);
		view.keydn(' ');
		Part p = f.getRootPart();
		boolean result;
		if (p.same(shouldbe)) {
			result = true;
		} else {
			p.dump();
			shouldbe.dump();
			result = false;
		}
		canvas.pop();
		assert(result);
	}

	public void testSpace2() {
//		System.out.println("test 2 (28-feb-2000)");
		// 6-dec-2000: not really the same bug as 28-feb-2000 but
		// when typing a space to change "xsubseteq y" into "x subseteq y",
		// the formula is OK but a NullPointerException is thrown in IdentifierView.splitme()
		// and moreover, the cursor part still is the "subseteq" Identifier (which
		// has been replaced by an OperatorSet).
		Formula f;
		Part shouldbe = new OperatorSet("subseteq",new Identifier("x"),new Identifier("y"));
		Identifier critical = new Identifier("xsubseteq");
		canvas.push(f = new Formula(new OperatorSpace(critical,new Identifier("y"))));
		canvas.setActiveFormula(0);
		FormulaView view = (FormulaView)canvas.getComponent(0);
		view.setCursorPart(critical,1);
		boolean result = true;
		try {
			view.keydn(' ');
		} catch (Exception exc) {
			exc.printStackTrace();
			result = false;
		}
		Part p = f.getRootPart();
		if (!p.same(shouldbe)) {
			p.dump();
			shouldbe.dump();
			result = false;
		}
		canvas.pop();
		assert(result);
	}

	public void testSpace3() {
//		System.out.println("test 3 (6-dec-2000)");
		// type "x y", set cursor at the left of the space, type another
		// space => crash
		Formula f;
		Part shouldbe = new OperatorSpace(new Identifier("x"),new Identifier("y"));
		Part critical = new OperatorSpace(new Identifier("x"),new Identifier("y"));
		canvas.push(f = new Formula(critical));
		canvas.setActiveFormula(0);
		FormulaView view = (FormulaView)canvas.getComponent(0);
		view.setCursorPart(critical,0);
		boolean result = true;
		try {
			view.keydn(' ');
		} catch (Exception exc) {
			exc.printStackTrace();
			result = false;
		}
		Part p = f.getRootPart();
		if (!p.same(shouldbe)) {
			p.dump();
			shouldbe.dump();
			result = false;
		}
		canvas.pop();
		assert(result);
	}

	public void testDot() {
//		System.out.println("test 4 (10-dec-2000)");
		// type "gve*C x" where "*" is the cursor, then type "."
		// Cause: OperatorSpace.recognizeOp() returned always
		// "true" in this case so the split of ".C" was not performed
		// since Identifier.splitme() thought that an identifier was
		// already recognized
		Formula f;
		Part shouldbe = new OperatorSpace(
			new OperatorDot(new Identifier("gve"),new Identifier("C")),
			new Identifier("x"));
		Part critical = new Identifier("gveC");
		canvas.push(f = new Formula(new OperatorSpace(critical,new Identifier("x"))));
		canvas.setActiveFormula(0);
		FormulaView view = (FormulaView)canvas.getComponent(0);
		view.setCursorPart(critical,3);
		boolean result = true;
		try {
			view.keydn('.');
		} catch (Exception exc) {
			exc.printStackTrace();
			result = false;
		}
		Part p = f.getRootPart();
		if (!p.same(shouldbe)) {
			p.dump();
			shouldbe.dump();
			result = false;
		}
		Part cp = view.getCursorPart();
		// PENDING: cursor should be at pos 0 of "C" identifier
		if (!(cp instanceof Identifier)) {
			System.out.println("Cursor not on OperatorDot");
			result = false;
		} else {
			Identifier cpIdent = (Identifier)cp;
			if (!cpIdent.getString().equals("C")) {
				System.out.println("Cursor not on C but on "+cpIdent.getString());
				result = false;
			}
			//// BUG IS IN InfixBinaryOp:96
			// Before: cursor is at "C", pos 0;
			// After: cursor is at OperatorDot
			int cpos = ((HasCursorPos)view.getCursorComp()).getCursorPos();
			if (cpos != 0) {
				System.out.println("CursorPos is "+cpos+", not 0");
				result = false;
			}
		}
		canvas.pop();
		assert(result);
	}
	
	public void test5() {
//		System.out.println("test 5 (26-jan-2001)");
		// move cursor with cursor keys over Uneditable part
		// => null pointer exception
		// at gve.calc.formula.UneditableView.setCursorPos(UneditableView.java:18)
		// Cause: the UneditableView probably has been replaced
		// by another component by the time that setCursorPos() is
		// called, so its parent is null.
		// Fix: replaced "FormulaView.get(this).setCursorComp(null);"
		// with "return;".
		Formula f;
//		canvas.newObject("gve.calc.formula.Stack");
		Stack stack = new Stack();
		canvas.push(f = new Formula(stack));
		boolean result = true;
		FormulaView view = (FormulaView)canvas.getComponent(0);
		view.setCursorPart(stack.elementAt(0));
		try {
			view.keydn(KeydnListener.Key_LEFT);
		} catch (NullPointerException exc) {
			exc.printStackTrace();
			result = false;
		}
		canvas.pop();
		assert(result);
	}

	public void test6() {
		// Type "a"; extend to "b,a" => result is "b ,a" where "b" and ",a" are two Identifiers!
//		System.out.println("test 6 (6-feb-2001)");
		boolean result = true;
		Identifier ident = new Identifier("a");
		Formula f;
		canvas.push(f = new Formula(ident));
		FormulaView view = (FormulaView)canvas.getComponent(0);
		view.setCursorPart(ident,HasCursorPos.Somewhere_LEFT);
		view.keydn('b');
		view.keydn(',');
		Part p = f.getRootPart();
		Part shouldbe = new OperatorComma(new Identifier("b"),new Identifier("a"));
		if (!p.same(shouldbe)) {
			p.dump();
			shouldbe.dump();
			result = false;
		}
		canvas.pop();
		assert(result);
	}

	public void test7() {
		// Type "a,b" and delete the comma
//		System.out.println("test 7 (15-feb-2001)");
		OperatorComma comma = new OperatorComma(new Identifier("a"),new Identifier("b"));
		Formula f;
		canvas.push(f = new Formula(comma));
		FormulaView view = (FormulaView)canvas.getComponent(0);
		view.setCursorPart(comma,HasCursorPos.Somewhere_MYRIGHT);
		view.keydn(KeydnListener.Key_BS);
		Part shouldbe = new Identifier("ab");
		Part root = f.getRootPart();
		boolean result = true;
		if (!root.same(shouldbe)) {
			root.dump();
			shouldbe.dump();
			result = false;
		}
		canvas.pop();
		assert(result);
	}

	public void test8() {
		// Type "nomt x" and repair to "not x"
//		System.out.println("test 8 (16-apr-2001)");

		Identifier nomt,iks;
		Part test = new OperatorSpace(nomt = new Identifier("nomt"),iks = new Identifier("x"));
		Formula f;
		canvas.push(f = new Formula(test));
		FormulaView view = (FormulaView)canvas.getComponent(0);
		view.setCursorPart(nomt,3);
		view.keydn(KeydnListener.Key_BS);
		view.keydn(KeydnListener.Key_RIGHT);
		view.keydn(KeydnListener.Key_RIGHT);
		Part shouldbe = new OperatorNot(new Identifier("x"));
		Part root = f.getRootPart();
		boolean result = true;
		if (!root.same(shouldbe)) {
			root.dump();
			shouldbe.dump();
			result = false;		
		}
		Part cp = view.getCursorPart();
		if (cp != root  &&  cp != iks) {
			System.out.println("bad cursor part "+view.getCursorPart());
			result = false;
		}
		canvas.pop();
		assert(result);
	}

	public void test9() {
		// Type "not x" and change into "nomt x"
//		System.out.println("test 9 (16-apr-2001)");

		Identifier nomt,iks;
		Part test = new OperatorNot(new Identifier("x"));
		Formula f;
		canvas.push(f = new Formula(test));
		FormulaView view = (FormulaView)canvas.getComponent(0);
		view.setCursorPart(test,2);
		view.keydn('m');
		Part shouldbe = new OperatorSpace(new Identifier("nomt"),new Identifier("x"));
		Part root = f.getRootPart();
		boolean result = true;
		if (!root.same(shouldbe)) {
			root.dump();
			shouldbe.dump();
			result = false;
		}
		canvas.pop();
		assert(result);
	}

	public void testForall() {
		// 3-Apr-2001
		// Type "forall y:a", change into "forall x:forall y:a"
		// => associates to "forall x: ((forall y):a) !
		Part test = new OperatorQuantor("forall",new OperatorColon(new Identifier("y"),new Identifier("a")));
		Formula f;
		canvas.push(f = new Formula(test));
		FormulaView view = (FormulaView)canvas.getComponent(0);
		view.setCursorPart(test,HasCursorPos.Somewhere_LEFT);
		String keys="forall x:";
		for (int i = 0; i < keys.length(); i++)
			view.keydn(keys.charAt(i));
		Part shouldbe  = new OperatorQuantor("forall",new OperatorColon(
			new Identifier("x"),
			new OperatorQuantor("forall",new OperatorColon(
				new Identifier("y"),new Identifier("a")
			))
		));
		Part root = f.getRootPart();
		boolean result = true;
		if (!root.same(shouldbe)) {
			root.dump();
			shouldbe.dump();
			result = false;
		}
		canvas.pop();
		assert(result);
	}

	public void testSemicolon() {
		// 4-May-2001
		// type "a;", then try to remove the semicolon => StringIndexOutOfBoundsException
		Part test = new OperatorSemicolon(new Identifier("a"));
		Formula f;
		canvas.push(f = new Formula(test));
		FormulaView view = (FormulaView)canvas.getComponent(0);
		view.setCursorPart(test,HasCursorPos.Somewhere_RIGHT);
		view.keydn(KeydnListener.Key_BS);
		Part root = f.getRootPart();
		boolean result = true;
		if (!root.same(new Identifier("a"))) {
			root.dump();
			result = false;
		}
		canvas.pop();
		assert(result);
	}

	public void testSpace() {
		// 17-May-2001
		// Type "Formula a" => space gets ignored!
		Part test = new Identifier("Formula");
		Formula f;
		canvas.push(f = new Formula(test));
		FormulaView view = (FormulaView)canvas.getComponent(0);
		view.setCursorPart(test,HasCursorPos.Somewhere_RIGHT);
		view.keydn(' ');
		view.keydn('a');
		Part root = f.getRootPart();
		Part shouldbe = new OperatorSpace(new Identifier("Formula"),new Identifier("a"));
		boolean result = true;
		if (!root.same(shouldbe)) {
			root.dump();
			shouldbe.dump();
			result = false;
		}
		canvas.pop();
		assert(result);
	}

	// Strange: catching exceptions doesn't work!
	public void testRandom() /*throws Exception*/ {
		String verbose = "";
		java.util.Random rnd = new java.util.Random();
		for (int i = 100; i >= 0; i--) {
			int key;
			key = rnd.nextInt(128);
			// Since exception catching doesn't work(!), print keys on every action
			if (verbose.length() < 30) System.out.println(verbose);
			else System.out.println(verbose.substring(verbose.length()-30));
//			try {
				if (key == 7) {
					canvas.keyTyped(new KeyEvent(canvas,KeyEvent.KEY_TYPED,0L,0,KeyEvent.VK_UNDEFINED,KeydnListener.Key_TAB));
					verbose += "TAB ";
				} else if (key == 6) {
					canvas.keyTyped(new KeyEvent(canvas,KeyEvent.KEY_TYPED,0L,KeyEvent.SHIFT_MASK,KeyEvent.VK_UNDEFINED,KeydnListener.Key_TAB));
					verbose += "SHTAB ";
				} else if (key>=32 && key<127) {
					verbose += "'"+(char)key+"' ";
					canvas.keyTyped(new KeyEvent(canvas,KeyEvent.KEY_TYPED,0L,0,KeyEvent.VK_UNDEFINED,(char)key));
				} else {
					int keycode;
					String keyname;
					switch (key) {
					case 0: keyname="UP"; keycode = KeyEvent.VK_UP; break;
					case 1: keyname="DOWN"; keycode = KeyEvent.VK_DOWN; break;
					case 2: keyname="LEFT"; keycode = KeyEvent.VK_LEFT; break;
					case 3: keyname="RIGHT"; keycode = KeyEvent.VK_RIGHT; break;
					case 4: keyname="ENTER"; keycode = KeyEvent.VK_ENTER; break;
					default: continue;
					}
					canvas.keyPressed(new KeyEvent(canvas,KeyEvent.KEY_PRESSED,0L,0,keycode));
					verbose += keyname+" ";
				}
/*			} catch (ArrayIndexOutOfBoundsException e) {
				System.out.println(verbose);
				throw e;
			}*/
		}
	}

	public void testQuote() {
		// 17-May-2001
		// type "ab" followed by a '"' => nothing happens
		// type "a+b" followed by a '"' => throws exception and nothing continues to happen
		Part b;
		Part test = new OperatorPlus(new Identifier("a"),b = new Identifier("b"));
		Formula f;
		canvas.push(f = new Formula(test));
		FormulaView view = (FormulaView)canvas.getComponent(0);
		view.setCursorPart(b,HasCursorPos.Somewhere_RIGHT);
		view.keydn('"');
		Part shouldbe = new OperatorPlus(new Identifier("a"),new OperatorSpace(
			new Identifier("b"),new StringPart("")));
		Part root = f.getRootPart();
		boolean result = true;
		if (!root.same(shouldbe)) {
			root.dump();
			shouldbe.dump();
			result = false;
		}
		canvas.pop();
		assert(result);
	}
}
